"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs = __importStar(require("fs"));
const path = __importStar(require("path"));
const axios_1 = __importDefault(require("axios"));
const system_1 = require("../utils/system");
const task_1 = require("../utils/task");
const time_1 = require("../utils/time");
const log_1 = __importDefault(require("../utils/log"));
const m3u8_1 = require("./m3u8");
const downloader_1 = __importStar(require("./downloader"));
const common_1 = require("../utils/common");
const file_concentrator_1 = require("./file_concentrator");
class ArchiveDownloader extends downloader_1.default {
    /**
     *
     * @param m3u8Path
     * @param config
     * @param config.threads 线程数量
     */
    constructor(m3u8Path, { threads, output, tempDir, key, verbose, retries, proxy, slice, format, cookies, headers, keep, noMerge, keepEncryptedChunks, chunkNamingStrategy, cliMode, } = {}) {
        super(m3u8Path, {
            threads: threads || 5,
            output,
            tempDir,
            key,
            verbose,
            retries,
            proxy,
            format,
            cookies,
            headers,
            keep,
            noMerge,
            keepEncryptedChunks,
            chunkNamingStrategy,
            cliMode,
        });
        this.downloadTasks = [];
        this.allDownloadTasks = [];
        // 使用 Object 的原因是使用数组，检索需要遍历数组 1/2 * n^2 次
        // 当有数千块的时候有那么一点点不可接受
        this.finishedFilenames = {};
        this.runningThreads = 0;
        this.isResumed = false; // 是否为恢复模式
        this.isDownloaded = false;
        if (slice) {
            this.sliceStart = (0, time_1.timeStringToSeconds)(slice.split("-")[0]);
            this.sliceEnd = (0, time_1.timeStringToSeconds)(slice.split("-")[1]);
        }
    }
    /**
     * Parse M3U8 Information
     */
    parse() {
        return __awaiter(this, void 0, void 0, function* () {
            // parse m3u8
            if (this.m3u8.encryptKeys.length > 0) {
                // Encrypted
                const key = this.m3u8.encryptKeys[0];
                if (key.startsWith("abematv-license")) {
                    log_1.default.info("Site comfirmed: AbemaTV.");
                    const parser = yield Promise.resolve().then(() => __importStar(require("./parsers/abema")));
                    parser.default.parse({
                        downloader: this,
                    });
                }
                else if (this.m3u8Path.includes("d22puzix29w08m")) {
                    log_1.default.info("Site comfirmed: Hibiki-Radio.");
                    const parser = yield Promise.resolve().then(() => __importStar(require("./parsers/hibiki")));
                    parser.default.parse({
                        downloader: this,
                    });
                }
                else {
                    log_1.default.warning(`Site is not supported by Minyami Core. Try common parser.`);
                    try {
                        const parser = yield Promise.resolve().then(() => __importStar(require("./parsers/common")));
                        yield parser.default.parse({
                            downloader: this,
                        });
                    }
                    catch (e) {
                        log_1.default.error("Aborted due to critical error.", e);
                        this.emit("critical-error", e);
                    }
                }
            }
            else {
                // Not encrypted
                if (this.m3u8Path.includes("dmc.nico")) {
                    // NicoNico
                    if (this.m3u8.m3u8Content.includes("#EXT-X-PLAYLIST-TYPE:VOD")) {
                        log_1.default.info("Site comfirmed: NicoVideo.");
                        try {
                            const parser = yield Promise.resolve().then(() => __importStar(require("./parsers/common")));
                            yield parser.default.parse({
                                downloader: this,
                            });
                        }
                        catch (e) {
                            log_1.default.error("Aborted due to critical error.", e);
                            this.emit("critical-error", e);
                        }
                    }
                    else {
                        log_1.default.info("Site comfirmed: NicoLive.");
                        const parser = yield Promise.resolve().then(() => __importStar(require("./parsers/nico")));
                        if (!this.key) {
                            log_1.default.info("请保持播放页面不要关闭");
                            log_1.default.info("Please do not close the video page.");
                            log_1.default.info(`Maybe you should get a audience token to get a better user experience.`);
                        }
                        if (this.threads > 10) {
                            log_1.default.warning(`High threads setting detected. Use at your own risk!`);
                        }
                        parser.default.parse({
                            downloader: this,
                        });
                        this.autoGenerateChunkList = false;
                    }
                }
                else if (this.m3u8Path.includes("googlevideo")) {
                    // YouTube
                    log_1.default.info("Site comfirmed: YouTube.");
                    const parser = yield Promise.resolve().then(() => __importStar(require("./parsers/youtube")));
                    parser.default.parse({
                        downloader: this,
                    });
                }
                else if (this.m3u8Path.includes("bcovlive")) {
                    // stagecrowd
                    log_1.default.info("Site comfirmed: Stagecrowd.");
                    try {
                        const siteParser = yield Promise.resolve().then(() => __importStar(require("./parsers/bcovlive")));
                        const parser = yield Promise.resolve().then(() => __importStar(require("./parsers/common")));
                        yield siteParser.default.parse({
                            downloader: this,
                        });
                        yield parser.default.parse({
                            downloader: this,
                        });
                    }
                    catch (e) {
                        log_1.default.error("Aborted due to critical error.", e);
                        this.emit("critical-error", e);
                    }
                }
                else {
                    log_1.default.warning(`Site is not supported by Minyami Core. Try common parser.`);
                    try {
                        const parser = yield Promise.resolve().then(() => __importStar(require("./parsers/common")));
                        yield parser.default.parse({
                            downloader: this,
                        });
                    }
                    catch (e) {
                        log_1.default.error("Aborted due to critical error.", e);
                        this.emit("critical-error", e);
                    }
                }
            }
        });
    }
    download() {
        return __awaiter(this, void 0, void 0, function* () {
            // Record start time to calculate speed.
            this.startedAt = new Date().valueOf();
            // Allocate temporary directory.
            this.tempPath = path.resolve(this.tempPath, "minyami_" + new Date().valueOf());
            if (!fs.existsSync(this.tempPath)) {
                fs.mkdirSync(this.tempPath);
            }
            if (this.cliMode) {
                process.on("SIGINT", () => __awaiter(this, void 0, void 0, function* () {
                    yield this.clean();
                    this.emit("finished");
                }));
            }
            try {
                yield this.parse();
                yield this.checkKeys();
            }
            catch (e) {
                log_1.default.error("Failed to parse M3U8 file.");
                log_1.default.debug(e);
                return;
            }
            log_1.default.info(`Start downloading with ${this.threads} thread(s).`);
            if (this.autoGenerateChunkList) {
                this.downloadTasks = this.m3u8.chunks.map((chunk, index) => {
                    const ext = (0, common_1.getFileExt)(chunk.url);
                    return {
                        url: chunk.url,
                        filename: this.onTaskOutputFileNaming(chunk, index),
                        retryCount: 0,
                        chunk,
                        id: index,
                    };
                });
            }
            if (this.sliceStart !== undefined && this.sliceEnd !== undefined) {
                const newChunkList = [];
                let nowTime = 0;
                for (const task of this.downloadTasks) {
                    if (nowTime > this.sliceEnd) {
                        break;
                    }
                    if ((0, downloader_1.isTaskGroup)(task)) {
                        const chunkGroupTotalLength = task.subTasks.reduce((prevLength, t) => prevLength + ((0, m3u8_1.isInitialChunk)(t.chunk) ? 0 : t.chunk.length), 0);
                        // 处理一组块
                        if (nowTime + chunkGroupTotalLength < this.sliceStart) {
                            // 加上整个块都还没有到开始时间
                            nowTime += chunkGroupTotalLength;
                            continue;
                        }
                        else {
                            // 组中至少有一个已经在时间范围内
                            const newChunkItem = {
                                actions: task.actions,
                                subTasks: [],
                                isFinished: false,
                                isNew: true,
                            };
                            for (const t of task.subTasks) {
                                if ((0, m3u8_1.isInitialChunk)(t.chunk)) {
                                    newChunkItem.subTasks.push(t);
                                    continue;
                                }
                                if (nowTime + t.chunk.length >= this.sliceStart &&
                                    nowTime + t.chunk.length < this.sliceEnd) {
                                    // 添加已经在时间范围内的块
                                    newChunkItem.subTasks.push(t);
                                    nowTime += t.chunk.length;
                                }
                                else {
                                    // 跳过时间范围外的块
                                    nowTime += t.chunk.length;
                                    continue;
                                }
                            }
                            newChunkList.push(newChunkItem);
                        }
                    }
                    else {
                        if ((0, m3u8_1.isInitialChunk)(task.chunk)) {
                            newChunkList.push(task);
                        }
                        else {
                            // 处理普通块
                            if (nowTime >= this.sliceStart) {
                                newChunkList.push(task);
                                nowTime += task.chunk.length;
                            }
                            else {
                                nowTime += task.chunk.length;
                            }
                        }
                    }
                }
                this.downloadTasks = newChunkList;
                // Mark skipped chunks as dropped
                const firstTask = (0, downloader_1.isTaskGroup)(this.downloadTasks[0])
                    ? this.downloadTasks[0].subTasks[0]
                    : this.downloadTasks[0];
                if (firstTask.id > 0) {
                    for (let i = 0; i < firstTask.id; i++) {
                        this.taskStatusRecord[i] = file_concentrator_1.TaskStatus.DROPPED;
                    }
                }
                else if (firstTask.chunk.isInitialChunk) {
                    // re-index chunks
                    if ((0, downloader_1.isTaskGroup)(this.downloadTasks[0])) {
                        // TODO: not supported
                        log_1.default.error("TaskGroup is not supported when initial segment is provided.");
                        return;
                    }
                    else {
                        this.downloadTasks.slice(1).forEach((task, index) => {
                            if (!(0, downloader_1.isTaskGroup)(task)) {
                                task.id = index;
                            }
                        });
                    }
                }
            }
            this.allDownloadTasks = this.downloadTasks.map((task) => {
                if ((0, downloader_1.isTaskGroup)(task)) {
                    return Object.assign(Object.assign({}, task), { subTasks: [...task.subTasks] });
                }
                else {
                    return task;
                }
            });
            this.totalChunksCount = 0;
            for (const chunk of this.downloadTasks) {
                this.totalChunksCount += (0, downloader_1.isTaskGroup)(chunk) ? chunk.subTasks.length : 1;
            }
            if (this.verbose) {
                setInterval(() => {
                    log_1.default.debug(`Now running threads: ${this.runningThreads}, finished chunks: ${this.finishedChunkCount}, total chunks: ${this.totalChunksCount}`);
                }, 3000);
            }
            this.checkQueue();
        });
    }
    /**
     * calculate ETA
     */
    getETA() {
        const usedTime = new Date().valueOf() - this.startedAt;
        const remainingTimeInSeconds = Math.round(((usedTime / this.finishedChunkCount) * this.totalChunksCount - usedTime) / 1000);
        if (remainingTimeInSeconds < 60) {
            return `${remainingTimeInSeconds}s`;
        }
        else if (remainingTimeInSeconds < 3600) {
            return `${Math.floor(remainingTimeInSeconds / 60)}m ${remainingTimeInSeconds % 60}s`;
        }
        else {
            return `${Math.floor(remainingTimeInSeconds / 3600)}h ${Math.floor((remainingTimeInSeconds % 3600) / 60)}m ${remainingTimeInSeconds % 60}s`;
        }
    }
    /**
     * Check task queue
     */
    checkQueue() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.downloadTasks.length > 0 && this.runningThreads < this.threads) {
                const firstTask = this.downloadTasks[0];
                let task;
                if ((0, downloader_1.isTaskGroup)(firstTask)) {
                    if (firstTask.actions && firstTask.isNew) {
                        log_1.default.debug(`Handle chunk actions for a new chunk group.`);
                        firstTask.isNew = false;
                        for (const action of firstTask.actions) {
                            yield this.handleChunkGroupAction(action);
                        }
                        this.checkQueue();
                        return;
                    }
                    if (firstTask.subTasks.length > 0) {
                        task = firstTask.subTasks.shift();
                        task.parentGroup = firstTask;
                        if (task.parentGroup.retryActions) {
                            for (const action of task.parentGroup.actions) {
                                yield this.handleChunkGroupAction(action);
                            }
                            task.parentGroup.retryActions = false;
                        }
                    }
                    else {
                        // All chunks finished in group
                        log_1.default.debug(`Skip a empty chunk group.`);
                        firstTask.isFinished = true;
                        this.downloadTasks.shift();
                        this.checkQueue();
                        return;
                    }
                }
                else {
                    task = this.downloadTasks.shift();
                    // this.chunks.shift();
                }
                this.runningThreads++;
                this.handleTask(task)
                    .then(() => {
                    this.runningThreads--;
                    const currentChunkInfo = {
                        taskname: task.filename,
                        finishedChunksCount: this.finishedChunkCount,
                        totalChunksCount: this.totalChunksCount,
                        chunkSpeed: this.calculateSpeedByChunk(),
                        ratioSpeed: this.calculateSpeedByRatio(),
                        eta: this.getETA(),
                    };
                    log_1.default.info(`Processing ${currentChunkInfo.taskname} finished. (${currentChunkInfo.finishedChunksCount} / ${this.totalChunksCount} or ${((currentChunkInfo.finishedChunksCount / currentChunkInfo.totalChunksCount) *
                        100).toFixed(2)}% | Avg Speed: ${currentChunkInfo.chunkSpeed} chunks/s or ${currentChunkInfo.ratioSpeed}x | ETA: ${currentChunkInfo.eta})`);
                    this.finishedFilenames[task.filename] = true;
                    if (!this.noMerge) {
                        this.fileConcentrator.addTasks([
                            {
                                filePath: task.chunk.isEncrypted
                                    ? path.resolve(this.tempPath, `./${task.filename}.decrypt`)
                                    : path.resolve(this.tempPath, `./${task.filename}`),
                                index: task.id,
                            },
                        ]);
                        this.taskStatusRecord[task.id] = file_concentrator_1.TaskStatus.DONE;
                    }
                    this.emit("chunk-downloaded", currentChunkInfo);
                    this.checkQueue();
                })
                    .catch((e) => {
                    this.emit("chunk-error", e, task.filename);
                    this.runningThreads--;
                    // 重试计数
                    task.retryCount = task.retryCount ? task.retryCount + 1 : 1;
                    if (task.parentGroup) {
                        if (task.parentGroup.isFinished) {
                            // Add a new group to the queue.
                            this.downloadTasks.unshift({
                                subTasks: [task],
                                actions: task.parentGroup.actions,
                                isFinished: false,
                                isNew: true,
                            });
                        }
                        else {
                            task.parentGroup.retryActions = true;
                            task.parentGroup.subTasks.unshift(task);
                        }
                    }
                    else if (this._internal_dropChunksInArchiveMode && task.retryCount >= this.retries) {
                        // 对于部分网站 放弃无法下载的分块
                        this.taskStatusRecord[task.id] = file_concentrator_1.TaskStatus.DROPPED;
                        this.finishedChunkCount++;
                        log_1.default.warning(`Processing ${task.filename} failed, max retries exceed, drop.`);
                    }
                    else {
                        this.downloadTasks.unshift(task);
                    }
                    this.checkQueue();
                });
                this.checkQueue();
            }
            if (this.downloadTasks.length === 0 &&
                this.totalChunksCount === this.finishedChunkCount &&
                this.runningThreads === 0) {
                if (this.isDownloaded) {
                    return;
                }
                this.isDownloaded = true;
                log_1.default.info("All chunks downloaded. Start merging chunks.");
                this.emit("downloaded");
                this.saveTask();
                if (this.noMerge) {
                    log_1.default.info("Skip merging. Please merge video chunks manually.");
                    log_1.default.info(`Temporary files are located at ${this.tempPath}`);
                    this.emit("finished");
                    return;
                }
                log_1.default.info("Merging chunks...");
                yield this.fileConcentrator.waitAllFilesWritten();
                if (!this.keepTemporaryFiles) {
                    log_1.default.info("End of merging.");
                    log_1.default.info("Starting cleaning temporary files.");
                    try {
                        yield (0, system_1.deleteEmptyDirectory)(this.tempPath);
                    }
                    catch (e) {
                        log_1.default.warning(`Fail to delete temporary files, please delete manually or execute "minyami --clean" later.`);
                    }
                }
                try {
                    (0, task_1.deleteTask)(this.m3u8Path.split("?")[0]);
                }
                catch (error) {
                    log_1.default.warning("Fail to parse previous tasks, ignored.");
                    log_1.default.warning(error.message);
                }
                const outputPaths = this.fileConcentrator.getOutputFilePaths();
                if (outputPaths.length === 1) {
                    log_1.default.info(`All finished. Please checkout your files at [${path.resolve(outputPaths[0])}]`);
                }
                else {
                    log_1.default.info(`All finished. Please checkout your files at ${outputPaths
                        .map((p) => `[${path.resolve(p)}]`)
                        .join(", ")}.`);
                }
                this.emit("finished");
            }
        });
    }
    resume(taskId) {
        return __awaiter(this, void 0, void 0, function* () {
            const previousTask = (0, task_1.getTask)(taskId.split("?")[0]);
            if (!previousTask) {
                log_1.default.error("Can't find a task to resume.");
                this.emit("critical-error", new Error("Can't find a task to resume."));
            }
            log_1.default.info("Previous task found. Resuming.");
            if (this.cliMode) {
                process.on("SIGINT", () => __awaiter(this, void 0, void 0, function* () {
                    yield this.clean();
                    this.emit("finished");
                }));
            }
            this.m3u8Path = taskId;
            // Resume status
            this.tempPath = previousTask.tempPath;
            this.outputPath = previousTask.outputPath;
            this.threads = previousTask.threads;
            this.cookies = previousTask.cookies;
            this.headers = previousTask.headers;
            this.key = previousTask.key;
            this.verbose = previousTask.verbose;
            this.startedAt = new Date().valueOf();
            this.totalChunksCount = previousTask.totalChunksCount - previousTask.finishedChunksCount;
            this.retries = previousTask.retries;
            this.timeout = previousTask.timeout;
            this.proxy = previousTask.proxy;
            this.allDownloadTasks = previousTask.allDownloadTasks;
            this.downloadTasks = previousTask.downloadTasks;
            this.finishedFilenames = previousTask.finishedFilenames;
            if (this.headers && Object.keys(this.headers).length > 0) {
                // Apply global custom headers
                axios_1.default.defaults.headers.common = Object.assign(Object.assign(Object.assign({}, axios_1.default.defaults.headers.common), (this.cookies ? { Cookie: this.cookies } : {})), this.headers);
            }
            // Load M3U8
            yield this.loadM3U8();
            try {
                yield this.parse();
                yield this.checkKeys();
            }
            catch (e) {
                log_1.default.error("Fail to parse M3U8 file.");
                log_1.default.debug(e);
                return;
            }
            this.isResumed = true;
            log_1.default.info(`Start downloading with ${this.threads} thread(s).`);
            this.checkQueue();
        });
    }
    /**
     * 退出前的清理工作
     */
    clean() {
        return __awaiter(this, void 0, void 0, function* () {
            log_1.default.info("Saving task status.");
            this.saveTask();
            log_1.default.info("Please wait.");
        });
    }
    saveTask() {
        const unfinishedTasks = [];
        this.allDownloadTasks.forEach((task) => {
            if ((0, downloader_1.isTaskGroup)(task)) {
                let allFinishedFlag = true;
                const unfinishedTasksInGroup = [];
                for (const subTask of task.subTasks) {
                    if (!this.finishedFilenames[subTask.filename]) {
                        allFinishedFlag = false;
                        unfinishedTasksInGroup.push(subTask);
                    }
                }
                if (!allFinishedFlag) {
                    // 组中块未全部完成 加入未完成列表
                    unfinishedTasks.push(Object.assign(Object.assign({}, task), { subTasks: unfinishedTasksInGroup }));
                }
            }
            else {
                if (!this.finishedFilenames[task.filename]) {
                    unfinishedTasks.push(task);
                }
            }
        });
        let unfinishedTaskCount = 0;
        for (const task of unfinishedTasks) {
            unfinishedTaskCount += (0, downloader_1.isTaskGroup)(task) ? task.subTasks.length : 1;
        }
        log_1.default.info(`Downloaded: ${this.finishedChunkCount}; Waiting for download: ${unfinishedTaskCount}`);
        try {
            (0, task_1.saveTask)({
                id: this.m3u8Path.split("?")[0],
                tempPath: this.tempPath,
                m3u8Path: this.m3u8Path,
                outputPath: this.outputPath,
                threads: this.threads,
                cookies: this.cookies,
                headers: this.headers,
                key: this.key,
                verbose: this.verbose,
                startedAt: this.startedAt,
                finishedChunksCount: this.totalChunksCount - unfinishedTaskCount,
                finishedChunkLength: this.finishedChunkLength,
                totalChunksCount: this.totalChunksCount,
                retries: this.retries,
                timeout: this.timeout,
                proxy: this.proxy,
                downloadTasks: unfinishedTasks,
                allDownloadTasks: this.allDownloadTasks,
                finishedFilenames: this.finishedFilenames,
            });
        }
        catch (error) {
            log_1.default.warning("Fail to parse previous tasks, ignored.");
            log_1.default.warning(error.message);
        }
    }
}
exports.default = ArchiveDownloader;
//# sourceMappingURL=archive.js.map