"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const media_1 = require("../../utils/media");
const log_1 = __importDefault(require("../../utils/log"));
const common_1 = require("../../utils/common");
class EncryptionKeyFetchError extends Error {
}
class Parser {
    static parse(_a) {
        return __awaiter(this, arguments, void 0, function* ({ downloader }) {
            downloader.setOnKeyUpdated((_b) => __awaiter(this, [_b], void 0, function* ({ keyUrls, explicitKeys, m3u8Url, saveEncryptionKey }) {
                const keys = {};
                // collect all key urls
                for (let i = 0; i <= keyUrls.length - 1; i++) {
                    keys[(0, common_1.buildFullUrl)(m3u8Url, keyUrls[i])] = explicitKeys[i] || "";
                }
                // download all keys
                let counter = 1;
                for (const url of Object.keys(keys)) {
                    log_1.default.info(`Downloading decrypt keys. (${counter} / ${Object.keys(keys).length})`);
                    if (explicitKeys[counter - 1]) {
                        saveEncryptionKey((0, common_1.buildFullUrl)(m3u8Url, url), explicitKeys[counter - 1]);
                        continue;
                    }
                    let retryCounter = downloader.retries;
                    while (retryCounter > 0) {
                        try {
                            const response = yield (0, media_1.requestRaw)(url, {
                                responseType: "arraybuffer",
                            });
                            const hexKey = Array.from(new Uint8Array(response.data))
                                .map((i) => (i.toString(16).length === 1 ? "0" + i.toString(16) : i.toString(16)))
                                .join("");
                            saveEncryptionKey((0, common_1.buildFullUrl)(m3u8Url, url), hexKey);
                            break;
                        }
                        catch (e) {
                            if (retryCounter > 1) {
                                retryCounter--;
                                log_1.default.debug(e);
                                log_1.default.info("Download decryption key failed, retry.");
                            }
                            else {
                                throw new EncryptionKeyFetchError("Max retries exceeded. Abort.");
                            }
                        }
                    }
                    counter++;
                }
            }));
            return {};
        });
    }
}
exports.default = Parser;
//# sourceMappingURL=common.js.map