"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const ReconnectingWebSocket = require("@eridanussora/reconnecting-websocket");
const WebSocket = require("ws");
const agent_1 = __importDefault(require("../../utils/agent"));
const ua_1 = __importDefault(require("../../constants/ua"));
class Parser {
    static parse({ downloader }) {
        if (!downloader.m3u8.m3u8Url) {
            throw new Error("Missing m3u8 url for Niconico.");
        }
        if (!downloader.key) {
            throw new Error("Missing token for Niconico.");
        }
        const proxyAgent = agent_1.default.getProxyAgentInstance();
        downloader.once("parsed", () => {
            const liveId = downloader.key.match(/(.+?)_/)[1];
            const isChannelLive = !liveId.startsWith("lv");
            const socketUrl = isChannelLive
                ? `wss://a.live2.nicovideo.jp/unama/wsapi/v2/watch/${liveId}/timeshift?audience_token=${downloader.key}`
                : `wss://a.live2.nicovideo.jp/wsapi/v2/watch/${liveId}/timeshift?audience_token=${downloader.key}`;
            const socket = new ReconnectingWebSocket(socketUrl, undefined, {
                WebSocket: WebSocket,
                clientOptions: {
                    headers: {
                        "User-Agent": ua_1.default.CHROME_DEFAULT_UA,
                    },
                    agent: proxyAgent ? proxyAgent : undefined,
                },
            });
            socket.addEventListener("message", (message) => {
                const parsedMessage = JSON.parse(message.data);
                // Send heartbeat packet to keep alive
                if (parsedMessage.type === "ping") {
                    socket.send(JSON.stringify({
                        type: "pong",
                        body: {},
                    }));
                    socket.send(JSON.stringify({
                        type: "keepSeat",
                    }));
                }
            });
        });
        return {};
    }
}
exports.default = Parser;
//# sourceMappingURL=nicolive.js.map